@@ -4,6 +4,8 @@ package Log::ProgramInfo;
 
 Log::ProgramInfo - log global info from a perl programs.
 
+=cut
+
 ### HISTORY ###################################################################
 # Version               Date            Developer        Comments
 # 0.1.1                 2015-04-02      John Macdonald   Initial release to CPAN
@@ -19,15 +21,17 @@ Log::ProgramInfo - log global info from a perl programs.
 #                                                        Log username(s) groupname(s)
 #                                                        Include identifying info in separator line
 #                                                        Make output more easily parsable
+# 0.1.7                 2015-05-20      John Macdonald   Add user-pluggable log routines
+#                                                        Move log parser from test to module.
 #
 
 =head1 VERSION
 
-Version 0.1.6
+Version 0.1.7
 
 =cut
 
-our $VERSION = '0.1.6';
+our $VERSION = '0.1.7';
 
 use feature qw(say);
 use Data::Dumper;
@@ -35,7 +39,7 @@ use FindBin;
 use Time::HiRes qw(time);
 use DateTime;
 use DateTime::Duration;
-use Carp;
+use Carp qw(carp croak cluck);
 use Fcntl qw(:flock);
 
 
@@ -114,6 +118,77 @@ use Fcntl qw(:flock);
         LOGPROGRAMINFO_EXT=EXT     string  sets the target extension
         LOGPROGRAMINFO_DATE=DATE   string  sets the target filename LOGDATE selector
 
+    Adding extra loggable information:
+        If you want to add your own classes of loggable info, there are a
+        few restrictions.
+
+        You define a logging extension routine using:
+
+            Log::ProgramInfo::add_extra_logger( \&my_logger );
+
+        Your logger routine should be defined as:
+
+            sub my_logger {
+                my $write_entry = shift;
+                $write_entry->( $key1, $value );
+                $write_entry->( $key1, $key2, $value );
+            }
+
+        The $write_entry function passed to my_logger must be called with
+        2 or 3 arguments.  The leading arguments are major (and minor if
+        desired) keys, the final one os the value for the key(s).
+
+        Try to keep the first key to 7 characters, and the second to 8 to
+        keep the log readable by humans.  (It will be parseable even if you
+        use longer keys.)
+
+    Parsing the log file:
+        The log file is designed to be easily parsed.
+
+        A log always starts with a line beginning with 8 hash marks in column
+        one (########) plus some identifying info.
+
+        The value lines are of the form:
+
+        key     : value
+        key1    : key2    : value
+
+        The first key is extended to at least 7 characters with blanks, the
+        second key (if any) is extended to at least 8 characters.  There is
+        always a separator of (space(colon)(space) between a key and the
+        following field.  (A key can be provided with leading spaces for making
+        the log more readable by humans - the readlog function in the test suite
+        will remove such spaces.)
+
+        Two subroutines are available to do this parsing for you:
+
+            my $firstonly = 0;
+            @logs = readlog( $filepath    [, $acceptsub] [, $firstonly] );
+            @logs = parselog( $filehandle [, $acceptsub] [, $firstonly] );
+
+            $logs = readlog( $filepath    [, $acceptsub ], 1 );
+            $logs = parselog( $filehandle [, $acceptsub ] ,1 );
+
+        The first parameter to each is either a string pathname (for readlog)
+        or an already opened readable file handle (for parselog).
+
+        If a subroutine reference arg $acceptsub is provided, each log that is
+        read will be passed to that sub reference.  If the acceptsub returns
+        true the log is retained, otherwise it is discarded.  If a trailing
+        (non-sub-ref) value is provided, it selects whether only the first
+        (acceptable) log found will be returned as a single hash reference, or
+        whether all of the (accepted) logs in the file will be returned as a
+        list of hash references.`
+
+        The hash reference for each accepted log contains the key/value (or
+        key1 => { key2/value pairs }) from that log.
+
+        Whenever a key (or key1/key2 pair) is seen multiple times, the value
+        is an array ref instead of a scalar.  This only happens with the
+        MODULE pairs (MODULE/NAME, MODULE/LOC, MODULE/VERSION), and the INC
+        key.  (User-provided keys are not currently permitted to use the same
+        key set multiple times.)
+
 =cut
 
 # preserve command line info
@@ -132,6 +207,79 @@ my $kill_caught;
 
 my ($uid, $gid);
 my %cache;
+my %modkeys = ( NAME => 1, VERSION => 1, LOC => 1 );
+
+sub readlog {
+    my $file = shift;
+    open my $fh, '<', $file or die "Cannot open $file: $!";
+    return parselog( $fh, @_ );
+}
+
+sub parselog {
+    my $fh = shift;
+    my $accept;
+    $accept = shift if ref($_[0]) eq 'CODE';
+    my $firstonly = shift;
+    my @logs;
+    while (my $log = parse1log( $fh )) {
+        next if $accept && ! $accept->($log);
+        return $log if $firstonly;
+        push @logs, $log;
+    }
+    return @logs;
+}
+
+sub parse1log {
+    my $fh = shift;
+    my $log;
+    while (my $line = <$fh>) {
+        return $log if $line =~ /^$/;
+        next if $line =~ /^########/;
+        chomp $line;
+        $log ||= {};
+        my @keys = split ': ', $line;
+        s/^\s*// for @keys;
+        s/\s*$// for @keys;
+        die "Unexpected syntax in log line: $line\n" unless scalar(@keys) >= 2;
+        my $val = pop @keys;
+        my $key = shift @keys;
+        if (scalar(@keys) == 0) {
+            if ($key eq 'INC') {
+                my $list = $log->{$key} ||= [];
+                push @$list, $val;
+            }
+            else {
+                die "repeated key: $key" if exists $log->{$key};
+                $log->{$key} = $val;
+            }
+        }
+        else {
+            my $key2 = shift @keys;
+            die "invalid nested key: {" . join( '}{', $key, $key2, @keys, $val ) . "}"
+                if scalar(@keys);
+            if ($key eq 'MODULE') {
+                die "Unknown MODULE key ($key2)" unless $modkeys{$key2};
+                my $list = $log->{$key}{$key2} ||= [];
+                push @$list, $val;
+            }
+            else {
+                die "repeated key: $key $key2" if exists $log->{$key}{$key2};
+                $log->{$key}{$key2} = $val;
+            }
+        }
+    }
+    return $log;
+}
+
+my @extra_loggers = ();
+
+sub add_extra_logger {
+    for my $logger (@_) {
+        croak "arg to extra_loggers is not a code ref: " . Dumper($logger)
+            unless ref $logger eq 'CODE';
+        push @extra_loggers, $logger;
+    }
+}
 
 sub groupmap {
     my $list = shift;
@@ -218,6 +366,20 @@ sub catch_sig {
     finish_log("Killed with signal: $signame");
 }
 
+sub log_entry {
+    my $logfh = shift;
+    if (@_ == 2 ) {
+        printf $logfh "%-7s : %s\n", @_;
+    } elsif (@_ == 3 ) {
+        printf $logfh "%-7s : %-8s : %s\n", @_;
+    } else {
+        my $msg = "log_entry needs 2 or 3 arguments, got "
+            . scalar(@_);
+        $msg .= ': (' . join( '), (', @_ ) . ')' if @_;
+        cluck $msg;
+    }
+}
+
 sub finish_log {
     return if $kill_caught++; # only write log once - first kill, or termination
     my $exit_status = shift;
@@ -263,15 +425,15 @@ sub finish_log {
         my $mod = show_modules();
         for my $key ( sort keys %$mod ) {
             my ( $ver, $loc ) = @{ $mod->{$key} };
-            say $logfh "MODULE  : NAME     : $key";
-            say $logfh "MODULE  : VERSION  : $ver";
-            say $logfh "MODULE  : LOC      : $loc";
+            log_entry( $logfh, MODULE => NAME    => $key );
+            log_entry( $logfh, MODULE => VERSION => $ver );
+            log_entry( $logfh, MODULE => LOC     => $loc );
         }
         for my $inc (@INC) {
-            say $logfh "INC     : $inc";
+            log_entry( $logfh, INC => $inc );
         }
 
-        printf $logfh ( "UNAME   : %-8s : %s", $_->[1], qx( uname $_->[0] ) )
+        log_entry( $logfh, UNAME => $_->[1], do { my $out = qx( uname $_->[0] ); chomp $out; $out } )
             for (
             [ -s => "System" ],
             [ -n => "Name" ],
@@ -279,22 +441,24 @@ sub finish_log {
             [ -v => "OSVer" ],
             [ -m => "Machine" ]
         );
-        say $logfh   "PERL    : ", $^X;
-        say $logfh   "PERLVer : ", $];
-        say $logfh   "User    : ", $uid;
-        say $logfh   "Group   : ", $gid;
-        say $logfh   "ProgDir : $FindBin::Bin";
-        say $logfh   "Program : $progbase";
-        say $logfh   "Version : ", ( $::VERSION // "(No VERSION)" );
-        say $logfh   "Args    : ", scalar(@args);
+        log_entry( $logfh, PERL    => $^X );
+        log_entry( $logfh, PERLVer => $] );
+        log_entry( $logfh, User    => $uid );
+        log_entry( $logfh, Group   => $gid );
+        log_entry( $logfh, ProgDir => $FindBin::Bin );
+        log_entry( $logfh, Program => $progbase );
+        log_entry( $logfh, Version => ( $::VERSION // "(No VERSION)" ) );
+        log_entry( $logfh, Args    => scalar(@args) );
         my $acnt = 0;
-        say $logfh   "  arg   : ", sprintf( "%-8d : %s", ++$acnt, $args[$acnt-1] ) for @args;
-        say $logfh   "Start   : ", $starttime->datetime(), ".", sprintf( "%03d", $starttime->millisecond );
-        say $logfh   "End     : ", $endtime->datetime(), ".", sprintf( "%03d", $endtime->millisecond );
+        log_entry( $logfh, "  arg" => sprintf("%8d", ++$acnt), $args[$acnt-1] ) for @args;
+        log_entry( $logfh, Start   => $starttime->datetime() . "." . sprintf( "%03d", $starttime->millisecond ) );
+        log_entry( $logfh, End     => $endtime->datetime()   . "." . sprintf( "%03d", $endtime->millisecond ) );
         my $dur = $endtime->subtract_datetime_absolute($starttime);
-        say $logfh   "Elapsed : ", $dur->delta_seconds, ".",
-                                    sprintf( "%03d", $dur->delta_nanoseconds/1_000_000);
-        say $logfh   "EndStat : ", $exit_status;
+        log_entry( $logfh, Elapsed => $dur->delta_seconds . "." .
+                                    sprintf( "%03d", $dur->delta_nanoseconds/1_000_000) );
+        log_entry( $logfh, EndStat => $exit_status );
+
+        $_->(sub { log_entry( $logfh, @_ ) }) for @extra_loggers;
 
         say $logfh "";
 
@@ -316,7 +480,7 @@ sub show_modules {
     # So let's convert it to this:
     # {
     #    ...
-    #    "Data/Dump.pm"
+    #    "Data::Dump"
     #        => [ "1.4.2",
     #             "/whatever/perl/lib/site_perl/5.18.1/Data/Dump.pm",
     #           ],
